from ..repository.repository import AbstractRepository


class AbstractRepositoryFactory:
    def get_repository(self, *args: object, **kwargs: object) -> AbstractRepository:
        self.check_type(args[0])
        return self._get_repository(*args, **kwargs)

    @staticmethod
    def check_type(_class):
        pass

    def _get_repository(self, *args, **kwargs):
        raise NotImplementedError()



from .state import State
from ..repository.storable import Storable


class StateMachine:
    def __init__(self, context, starting_state, states):
        self.initial_state = starting_state
        self.current_state = starting_state
        self.states = states
        self.context: Storable = context

    def next(self, trigger):
        self.current_state = self.current_state.get_transition_state(trigger)

    def __str__(self):
        return str(self.current_state)

    def to_dict(self):
        def get_all_connections(starting_state, arr=None, traversed=None):
            if traversed is None:
                traversed = []
            if arr is None:
                arr = []
            if starting_state not in traversed:
                for trigger, allowed_state in starting_state.next_states.items():
                    arr.append((starting_state.name, trigger, allowed_state.name))
                    traversed.append(starting_state)
                    get_all_connections(allowed_state, arr=arr, traversed=traversed)

            return arr

        states = [state.name for state in self.states]
        connections = get_all_connections(self.initial_state)

        return {
            'states': states,
            'connections': connections,
            'current_state': self.current_state.name,
            'starting_state': self.initial_state.name
        }

    @classmethod
    def from_dict(cls, data):
        states = dict([(name, State(name)) for name in data['states']])

        for (from_state_name, trigger, to_state_name) in data['connections']:
            from_state = states[from_state_name]
            to_state = states[to_state_name]
            from_state.add_allowed_transition(trigger, to_state)

        current_state = states[data['current_state']]
        instance = cls(None, current_state, states.values())
        instance.initial_state = states[data['starting_state']]

        return instance



from datetime import datetime
from typing import List, Callable

from ..event_queue import EventQueue
from ..events import DomainEvent
from ..events import Request
from ..events.error_response import ErrorResponse
from ..events.event import Event
from ..repository.repository_factory import AbstractRepositoryFactory
from ..topic.topic_factory import TopicFactory


class Handler:
    correlation_id: str
    event_queue: EventQueue
    topic_factory: TopicFactory
    repository_factory: AbstractRepositoryFactory
    return_address: str
    name: str
    current_event: Event
    initial_event: Event
    should_deactivate: bool

    def __init__(self, correlation_id):
        self.correlation_id = correlation_id
        self.running: bool = True
        self.key_readers = []
        self.initial_event = None
        self.should_deactivate = True

    @classmethod
    def handles(cls):
        return dict([(getattr(method, 'handles'), method) for method in cls.__dict__.values() if
                     hasattr(method, 'handles')])

    @classmethod
    def started_by(cls) -> List[str]:
        return list(map(lambda method: method.handles, filter(lambda method: hasattr(method, 'starts_handler'),
                                                              filter(lambda method: hasattr(method, 'handles'),
                                                                     list(cls.__dict__.values())))))

    @classmethod
    def get_teardown_method(cls) -> Callable:
        for method in cls.__dict__.values():
            if hasattr(method, 'is_teardown'):
                return method

        return None

    @staticmethod
    def subscribes_for():
        return []

    def send(self, event):
        if issubclass(type(event), Request):
            event.return_address = f'isc://{self.name}'
            event.system_entry = str(datetime.now())
            event.correlation_id = self.correlation_id
        elif issubclass(type(event), DomainEvent):
            event.system_entry = str(datetime.utcnow())
            event.correlation_id = self.correlation_id
            event.return_address = f'isc://{self.name}'
            event.origin = f'isc://{self.name}'

        self.event_queue.add_event(event)

    def respond_with_success(self, event, correlation_id=None):
        event.destination = self.return_address
        event.origin = f'isc://{self.name}'
        event.system_entry = self.current_event.system_entry
        event.started_handling = self.current_event.started_handling
        event.handling_stopped = str(datetime.now())
        event.correlation_id = self.current_event.correlation_id if not correlation_id else correlation_id
        self.event_queue.add_event(event)

    def stop_with_success(self, event, correlation_id=None):
        self.respond_with_success(event, correlation_id=correlation_id)
        self.running = False

    def respond_with_error(self, event, message=None, status=500):
        event.status = status
        event.message = message
        event.destination = self.return_address
        event.origin = f'isc://{self.name}'
        event.system_entry = self.current_event.system_entry
        event.started_handling = self.current_event.started_handling
        event.handling_stopped = str(datetime.now())
        event.correlation_id = self.current_event.correlation_id
        self.event_queue.add_event(event)

    def stop_with_error(self, event, message=None, status=500):
        self.respond_with_error(event, message, status=status)
        self.running = False

    def respond_with_success_to_initial_event(self, response):
        response.destination = self.initial_event.return_address
        response.correlation_id = self.initial_event.correlation_id
        response.system_entry = self.current_event.system_entry
        response.started_handling = self.current_event.started_handling
        response.handling_stopped = str(datetime.now())
        response.origin = f'isc://{self.name}'
        self.event_queue.add_event(response)

    def respond_with_error_to_initial_event(self, message='', status=''):
        response = ErrorResponse()
        response.status = status
        response.message = message
        response.payload = None
        response.destination = self.initial_event.return_address
        response.origin = f'isc://{self.name}'
        response.system_entry = self.current_event.system_entry
        response.started_handling = self.current_event.started_handling
        response.handling_stopped = str(datetime.now())
        response.correlation_id = self.initial_event.correlation_id
        self.event_queue.add_event(response)

    def stop(self):
        self.running = False



from ..repository.memory_repository import MemoryRepository
from ..repository.repository_factory import AbstractRepositoryFactory


class MemoryRepositoryFactory(AbstractRepositoryFactory):
    def __init__(self):
        super().__init__()
        self.repositories = {}

    def _get_repository(self, for_class):
        if for_class in self.repositories:
            return self.repositories[for_class]

        repository = MemoryRepository(for_class)
        self.repositories[for_class] = repository

        return repository



from ..repository.repository_extension.extension_bootstapper import ExtensionBootstrapper
from ..repository.mongo_repository import MongoRepository
from ..repository.repository_factory import AbstractRepositoryFactory


class MongoRepositoryFactory(AbstractRepositoryFactory):
    def __init__(self, mongo_client, extension_bootstrapper, mongo_database="emcai_database"):
        super().__init__()
        self.client = mongo_client
        self.extension_bootstrapper: ExtensionBootstrapper = extension_bootstrapper
        self.database = mongo_client[mongo_database]
        self.database_name = mongo_database
        self.repositories = {}

    def _get_repository(self, for_class):
        if for_class in self.repositories:
            return self.repositories[for_class]

        collection_name = for_class.__name__.lower() + '_collection'
        collection = self.database[collection_name]

        repository = MongoRepository(for_class, collection)

        for extension in self.extension_bootstrapper.get_extensions():
            if getattr(extension, 'extension_for') == for_class:
                for _, method in extension.__dict__.items():
                    if hasattr(method, 'include_into_extension'):
                        self._bind(repository, method)

        self.repositories[for_class] = repository

        return repository

    def _bind(self, instance, func):
        bound_method = func.__get__(instance, instance.__class__)
        setattr(instance, func.__name__, bound_method)

        return bound_method



from threading import Lock

from .events.event import Event


class EventQueue:
    def __init__(self):
        self.queue = []
        self.lock = Lock()

    def add_event(self, event):
        assert issubclass(type(event), Event), 'Only events shall be submitted to queue'

        with self.lock:
            self.queue.append(event)

    def get_event(self):
        return_value = None

        if len(self) != 0:
            with self.lock:
                return_value = self.queue[0]
                self.queue = self.queue[1:]

        return return_value

    def __len__(self):
        return len(self.queue)



from typing import Callable

from ..repository.storable import Storable
from .state import State


class StateConfigurator:
    def __init__(self, state, state_machine_builder):
        self.state: State = state
        self.permitted_states = {}
        self.state_machine_builder = state_machine_builder
        self.context = state_machine_builder.context

    def permit(self, trigger, name):
        assert trigger not in self.permitted_states.keys(), 'Only one state transition permitted per trigger'
        assert name in self.state_machine_builder.states, 'State machine does not contain state for ' + name
        self.state.add_allowed_transition(trigger, self.state_machine_builder.states[name])

        return self

    def on_transition_to(self, to_state: str, a_callable: Callable[[Storable], None]):
        assert to_state in self.state_machine_builder.states, 'State is not configurable'
        self.state.on_transition_to(self.state_machine_builder.states[to_state], a_callable)

    def ignore(self, trigger):
        self.state.add_allowed_transition(trigger, self.state)



from ..subscriber import SubscriberType
from .topic import Topic


class TopicFactory:
    def __init__(self, event_queue):
        self.event_queue = event_queue
        self.map = dict()

    def get_topic(self, topic_identifier) -> Topic:
        if topic_identifier in self.map:
            return self.map[topic_identifier]

        instance = Topic(self.event_queue, topic_identifier)
        self.map[topic_identifier] = instance
        return instance

    def remove_client(self, client_id):
        for (topic_identifier, topic) in self.map.items():
            subscribers_in_topic = list(
                filter(lambda subscriber: subscriber.type == SubscriberType.CLIENT and client_id in subscriber.address,
                       topic.subscribers))

            for sub in subscribers_in_topic:
                topic.remove_subscriber(sub)



from ..repository.storable import Storable
from .state import State
from .state_configurator import StateConfigurator
from .state_machine import StateMachine


class StateMachineBuilder:
    def __init__(self, context: Storable):
        self.starting_state = None
        self.states = {}
        self.context = context

    def add_state(self, *args):
        for name in args:
            assert name not in self.states
        for name in args:
            self.states[name] = State(name, context=self.context)

    def configure(self, name):
        assert name in self.states

        return StateConfigurator(self.states[name], self)

    def set_initial_state(self, name):
        assert name in self.states

        self.starting_state = self.states[name]

        return self

    def build(self):
        assert self.starting_state, 'No starting state defined!'

        return StateMachine(self.context, self.starting_state, self.states.values())



from .event import Event


class Notification(Event):
    def __init__(self, payload):
        super().__init__()
        self.payload = payload
        self.client_id = None

    @staticmethod
    def get_type():
        raise NotImplementedError()



from datetime import datetime


class Client:
    def __init__(self, client_id, websocket):
        self.client_id = client_id
        self.websocket = websocket
        self.connected_at = datetime.now()
        self.last_heartbeat = datetime.now()



import re
from copy import deepcopy
from typing import List, Type, Dict

from ..events.event import Event
from .handler import Handler


class HandlerBootstrapper:
    event_to_handler_map: 'Dict[Type[Event], List[Type[Handler]]]'
    handler_list: 'List[Type[Handler]]'
    service_name_to_class_map: 'Dict[str, Type[Handler]]'

    def __init__(self, handler_list: List[Type[Handler]]):
        self.handler_list = handler_list
        self.event_to_handler_map = {}
        self.service_name_to_class_map = dict([(self.get_service_name_from_class(
            handler_class), handler_class) for handler_class in self.handler_list])

        for handler in self.handler_list:
            for key, value in handler.__dict__.items():
                if hasattr(value, 'handles'):
                    event_type = getattr(value, 'handles')
                    if event_type in self.event_to_handler_map:
                        self.event_to_handler_map[event_type].append(handler)
                    else:
                        self.event_to_handler_map[event_type] = [handler]

    @staticmethod
    def get_service_name_from_class(handler_class: 'Type[Handler]') -> str:
        return '-'.join(re.findall(r'[A-Z][a-z]+', handler_class.__name__)).lower()

    def get_handlers(self) -> 'List[Type[Handler]]':
        return deepcopy(self.handler_list)

    def get_handler_for(self, event_type: 'Type[Event]') -> 'Type[Handler]':
        return self.event_to_handler_map[event_type][0] if event_type in self.event_to_handler_map else None

    def get_handlers_for_notification(self, notification_type) -> 'List[Type[Handler]]':
        handlers = list(
            filter(
                lambda handler: notification_type in handler.handles().keys(), self.handler_list
            )
        )
        return handlers

    def get_handler_class_by_service_name(self, service_name) -> 'Type[Handler]':
        return self.service_name_to_class_map[service_name] if service_name in self.service_name_to_class_map else None



import asyncio
import ssl
from datetime import datetime
from uuid import uuid4

import websockets

from .exit_state import ExitState
from . import serializer
from .client import Client
from .events.client_disconnected import ClientDisconnectedNotification
from .topic.topic_factory import TopicFactory


class WebsocketServer:
    def __init__(self, correlation_store, event_dispatcher, client_store, event_queue, exception_handler,
                 topic_factory, exit_state: ExitState):
        self.correlation_store = correlation_store
        self.event_dispatcher = event_dispatcher
        self.client_store = client_store
        self.event_queue = event_queue
        self.exception_handler = exception_handler
        self.topic_factory: TopicFactory = topic_factory
        self.exit_state: ExitState = exit_state
        self.number_of_incoming_events = 0

    async def send_heartbeat(self):
        while True:
            to_disconnect = []
            for client_id in self.client_store.keys():
                if not self.should_disconnect(client_id):
                    await self.send_to(client_id, 'HEARTBEAT')
                else:
                    to_disconnect.append(client_id)

            for client_id in to_disconnect:
                await self.unregister(client_id)

            await asyncio.sleep(5)

    async def register(self, websocket):
        client_id = str(uuid4())
        self.client_store.add(client_id, Client(client_id, websocket))
        await websocket.send('REGISTERED')
        return client_id

    async def unregister(self, client_id):
        client = self.client_store.get(client_id)
        await client.websocket.close()

        self.client_store.remove(client_id)

    def should_disconnect(self, client_id):
        client = self.client_store.get(client_id)
        since_heartbeat = (datetime.now() - client.last_heartbeat).seconds

        return True if since_heartbeat >= 5 or client.websocket.closed else False

    def got_client_heartbeat(self, client_id):
        client = self.client_store.get(client_id)
        client.last_heartbeat = datetime.now()

    async def websocket_connection_handler(self, websocket, path):
        client_id = await self.register(websocket)
        started_to_handle = set()

        try:
            while not self.exit_state.should_exit:
                correlation_id = str(uuid4())
                self.correlation_store.add(correlation_id, client_id)
                started_to_handle.add(correlation_id)

                message = await websocket.recv()

                if message == 'HEARTBEAT':
                    self.got_client_heartbeat(client_id)
                else:
                    try:
                        event = serializer.deserialize(
                            message, correlation_id, client_id)
                        event.return_address = f'remote://websocket/{client_id}'
                        self.event_queue.add_event(event)
                        self.number_of_incoming_events += 1
                    except Exception as e:
                        event = type('evt', (object,), dict(system_entry=str(
                            datetime.now()), destination=f'remote://websocket/{client_id}'))()

                        self.exception_handler.handle_exception(
                            e, event)
        except Exception as exception:
            print(
                f'client {client_id} suddenly disconnected. Reason: {type(exception).__name__} -> {exception}')
            self.client_store.remove(client_id)
            self.topic_factory.remove_client(client_id)
            self.topic_factory.get_topic('server_notifications').publish(ClientDisconnectedNotification(client_id),
                                                                         str(uuid4()))

    async def send_to(self, client_id, contents):
        client = self.client_store.get(client_id)
        if client:
            await client.websocket.send(contents)

    def run(self, address: str, port: int, ssl_context: ssl.SSLContext = None):
        start_server = websockets.serve(
            self.websocket_connection_handler, address, port, ssl=ssl_context)

        event_loop = asyncio.get_event_loop()
        heartbeat_task = event_loop.create_task(self.send_heartbeat())
        event_coroutine = event_loop.create_task(
            self.event_dispatcher.dispatch())

        print(f'Running on {"wss" if ssl_context else "ws"}://{address}:{port}')
        event_loop.run_until_complete(start_server)
        event_loop.run_until_complete(heartbeat_task)
        event_loop.run_until_complete(event_coroutine)


from .event import Event


class Response(Event):
    def __init__(self, status=200, message='OK', payload=None):
        super().__init__()
        self.status = status
        self.message = message
        self.payload = payload
        self.origin = None

    @staticmethod
    def get_type() -> str:
        raise NotImplementedError()



import datetime
from queue import Queue
from _queue import Empty
from threading import Thread
from time import sleep
from typing import Callable

from .handler.handler import Handler
from .repository.repository_factory import AbstractRepositoryFactory

from .event_queue import EventQueue
from .events.event import Event


class ExecutorThread(Thread):
    def __init__(self, handler_class, correlation_id, event_queue, exception_handler, topic_factory, repository_factory,
                 name, service_name, stopped, protected_check: 'Callable[[Event], bool]' = None,
                 public_check: 'Callable[[Event], bool]' = None):
        super().__init__()

        self.handler_class = handler_class
        self.correlation_id = correlation_id
        self.started_at = None
        self.event_queue: EventQueue = event_queue
        self.topic_factory = topic_factory
        self.exception_handler = exception_handler
        self.repository_factory = repository_factory
        self.execute_queue = Queue()
        self.name = name
        self.service_name = service_name
        self.stopped = stopped
        self.inactive_for = 0
        self._init_handler()

        def pass_all_event(event):
            return True

        self.protected_check: 'Callable[[Event, Callable, AbstractRepositoryFactory], bool]' = protected_check if protected_check else pass_all_event
        self.public_check: 'Callable[[Event, Callable, AbstractRepositoryFactory], bool]' = public_check if public_check else pass_all_event

    def _init_handler(self):
        self.handler_instance: Handler = object.__new__(self.handler_class)
        self.handler_instance.event_queue = self.event_queue
        self.handler_instance.repository_factory = self.repository_factory
        self.handler_instance.topic_factory = self.topic_factory
        self.handler_instance.correlation_id = self.correlation_id
        self.handler_instance.key_readers = []
        self.handler_instance.running = True
        self.handler_instance.name = self.name
        self.handler_instance.initial_event = None
        self.handler_instance.should_deactivate = True

        self.handler_instance.__init__()

    def run(self):
        self.started_at = datetime.datetime.now()

        while self.handler_instance.running:
            try:
                method, args = self.execute_queue.get_nowait()
                event = args[0]
                if not self.handler_instance.initial_event:
                    self.handler_instance.initial_event = event
                event.started_handling = str(datetime.datetime.now())
                self.handler_instance.current_event = event
                self.handler_instance.return_address = event.return_address
                try:
                    method(self.handler_instance, *args)
                    self.inactive_for = 0
                except Exception as exception:
                    self.exception_handler.handle_exception(
                        exception, event, use_return=True)
                    print(exception)
                    self.handler_instance.running = False
                    break
            except Empty as e:
                sleep(0.1)
                self.inactive_for += 0.1

                if self.inactive_for >= 10 and self.handler_instance.should_deactivate:
                    tear_down_method = self.handler_instance.get_teardown_method()
                    if tear_down_method:
                        bound_method = tear_down_method.__get__(self.handler_instance, self.handler_instance.__class__)
                        bound_method()

                    self.handler_instance.running = False
                    print(f'Stopping {self.service_name} due to belonged inactivity!')
                pass

        self.stopped(self.service_name, self.name)

    def running_for(self):
        return datetime.datetime.now() - self.started_at

    def process_event(self, *args):
        event = args[0]
        assert event, 'events must be of type events, not none'
        if event.get_type() in self.handler_instance.handles():
            handler_method = self.handler_instance.handles()[event.get_type()]
            if not self.__check_protected(handler_method, event) or not self.__check_only_public(handler_method, event):
                self.exception_handler.handle_exception(
                    Exception('Unauthorized access!'), event, status=400, use_return=True)
                self.handler_instance.running = False

            else:
                self.handler_instance.correlation_id = event.correlation_id
                self.execute_queue.put([handler_method, [*args]])

        else:
            raise Exception(f'This handler does not handle {event.get_type()}')

    def __check_protected(self, handler_method, event):
        if hasattr(handler_method, 'protected') and getattr(handler_method, 'protected'):
            return self.protected_check(event, handler_method, self.repository_factory)

        return True

    def __check_only_public(self, handler_method, event):
        return self.public_check(event, handler_method, self.repository_factory)

    def __str__(self):
        return f'<ExecutorThread started_at={self.started_at} service_name={self.service_name} />'



from .store import Store


class ClientStore(Store):
    def __init__(self):
        super().__init__()
        self.map = {}

    def _add(self, client_id, client):
        self.map[client_id] = client

    def _get(self, client_id):
        return self.map[client_id] if client_id in self.map.keys() else None

    def _remove(self, client_id):
        if client_id in self.map:
            del self.map[client_id]

    def _contains(self, client_id):
        return client_id in self.map

    def keys(self):
        with self.lock:
            return self.map.keys()



from uuid import uuid4


class Storable:
    def __init__(self):
        self.id = str(uuid4())

    def get_id(self):
        return self.id

    def to_dict(self):
        raise NotImplementedError()

    def __eq__(self, that):
        return self.get_id() == that.get_id()

    def __str__(self):
        s = ''
        for prop, value in vars(self).items():
            s += f'{str(prop)}={str(value)} '

        return f'<{self.__class__.__name__} {s} />'



from datetime import datetime

from ..events.domain_event import DomainEvent
from ..events.error_response import ErrorResponse


class ExceptionHandler:
    def __init__(self, event_queue):
        self.event_queue = event_queue

    def handle_exception(self, exception, event, status=500, use_return=False):
        try:
            message = f'{type(exception).__name__}: {exception}'
            if not event.destination and not use_return and not event.return_address:
                raise Exception('Event of type:', event.get_type(), 'has no destination')
            print(message)
            error_response = ErrorResponse(message=message, status=status)
            error_response.destination = event.return_address if use_return else event.destination
            error_response.system_entry = event.system_entry
            error_response.started_handling = event.started_handling if hasattr(event, 'started_handling') else str(
                datetime.now())
            error_response.stopped_handling = event.stopped_handling if hasattr(event, 'stopped_handling') else str(
                datetime.now())
            error_response.correlation_id = event.correlation_id
            if not issubclass(type(event), DomainEvent):
                self.event_queue.add_event(error_response)
        except Exception as e:
            print('Error building exception', e)



from copy import deepcopy
from typing import Type, List

from ...repository.repository_extension.extension import RepositoryExtension


class ExtensionBootstrapper:
    extension_list: 'List[Type[RepositoryExtension]]'

    def __init__(self, extension_list: 'List[Type[RepositoryExtension]]'):
        self.extension_list = deepcopy(extension_list)

    def get_extensions(self) -> 'List[Type[RepositoryExtension]]':
        return deepcopy(self.extension_list)



from .domain_event import DomainEvent
from .notification import Notification
from .request import Request
from .response import Response

__all__ = [
    'DomainEvent', 'Request', 'Response', 'Notification'
]


from .event import Event


class Request(Event):
    def __init__(self, event_type, payload=None, headers=None):
        super().__init__()

        if headers is None:
            headers = {}
        if payload is None:
            payload = {}

        self.payload = payload
        self.headers = headers
        self.type = event_type
        self.client_id = None

    def get_type(self):
        return self.type



from .store import Store


class CorrelationStore(Store):
    def __init__(self):
        super().__init__()
        self.map = {}

    def _add(self, correlation_id, client_id):
        self.map[correlation_id] = client_id

    def _get(self, correlation_id):
        return self.map[correlation_id] if correlation_id in self.map else None

    def _remove(self, correlation_id):
        if correlation_id in self.map:
            del self.map[correlation_id]

    def _contains(self, correlation_id):
        return correlation_id in self.map



from .notification import Notification


class ShutdownRequestedNotification(Notification):
    def __init__(self):
        super().__init__(payload={})

    @staticmethod
    def get_type():
        return 'notification.shutdown_requested'



import datetime
from copy import deepcopy

from ..subscriber import Subscriber


class Topic:
    def __init__(self, event_queue, topic_identifier):
        self.subscribers = set()
        self.event_queue = event_queue
        self.name: str = topic_identifier

    def add_subscriber(self, subscriber):
        self.subscribers.add(subscriber)

    def remove_subscriber_by_client_id(self, client_id):
        self.remove_subscriber(Subscriber.client_from_id(client_id))

    def remove_subscribers(self):
        subs = list(self.subscribers)
        for subscriber in subs:
            self.remove_subscriber(subscriber)

    def remove_subscriber(self, subscriber):
        if subscriber in self.subscribers:
            self.subscribers.remove(subscriber)

    def publish(self, notification, correlation_id):
        if len(self.subscribers) == 0:
            print(
                f'Notification {notification.get_type()} was not published, '
                f'because there are no subscribers on this ({self.name}) topic')
            return

        for subscriber in self.subscribers:
            copied_notification = deepcopy(notification)
            copied_notification.destination = subscriber.address
            copied_notification.correlation_id = correlation_id
            copied_notification.system_entry = str(datetime.datetime.now())
            self.event_queue.add_event(copied_notification)



from .notification import Notification


class DisconnectedNotification(Notification):
    def __init__(self, reason='Server shutdown!'):
        super().__init__(payload={
            'reason': reason
        })

    @staticmethod
    def get_type():
        return 'notification.disconnected'



def repository_extension(for_class):
    def sub_decorator(actual_class):
        actual_class.extension_for = for_class

        return actual_class

    return sub_decorator


def include_method(method):
    method.include_into_extension = True
    return method



import _thread
import datetime
import os
import signal
from time import sleep
from uuid import uuid4

from .event_queue import EventQueue
from .events.server_disconnect import DisconnectedNotification
from .events.shutdown_requested import ShutdownRequestedNotification
from .store.client_store import ClientStore
from .store.thread_store import ThreadStore
from .topic.topic import Topic


class ExitState:
    should_exit: bool

    def __init__(self, server_notifications_topic: Topic, client_store: ClientStore, thread_store: ThreadStore,
                 event_queue: EventQueue):
        self.should_exit = False
        self.client_store = client_store
        self.thread_store = thread_store
        self.server_notifications_topic = server_notifications_topic
        self.event_queue = event_queue

        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)
        signal.signal(signal.SIGUSR1, self.actual_exit)

    def exit_gracefully(self, signum, frame):
        def actual_exit():
            self.server_notifications_topic.publish(ShutdownRequestedNotification(), str(uuid4()))
            self.send_disconnect_message()
            sleep(2)
            self.should_exit = True
            for dictionary in self.thread_store.map.values():
                for executor_thread in list(dictionary.values()):
                    if executor_thread.handler_instance:
                        executor_thread.handler_instance.running = False
            os.kill(os.getpid(), signal.SIGUSR1)

        _thread.start_new_thread(actual_exit, ())

    def actual_exit(self, signum, frame):
        exit(1)

    def send_disconnect_message(self):
        for client_id in self.client_store.map.keys():
            print(client_id)
            event = DisconnectedNotification()
            event.destination = f'remote://websocket/{client_id}'
            event.correlation_id = str(uuid4())
            event.system_entry = str(datetime.datetime.now())
            self.event_queue.add_event(event)



class Event:
    def __init__(self):
        self.event_id = str(__import__('uuid').uuid4())
        self.correlation_id = None
        self.destination = None
        self.return_address = None
        self.system_entry = None
        self.system_leave = None
        self.retry_count = 0



import datetime
import threading
from uuid import uuid4

from .event import Event


class DomainEvent(Event):
    def __init__(self, payload):
        super().__init__()
        self.payload = payload

    @staticmethod
    def get_type():
        raise NotImplementedError()

    def raise_event(self):
        current_thread = threading.current_thread()
        self.system_entry = str(datetime.datetime.now())
        self.correlation_id = str(uuid4())
        current_thread.event_queue.add_event(self)



from copy import deepcopy

from ..repository.repository import AbstractRepository


class MemoryRepository(AbstractRepository):
    def __init__(self, for_class):
        super().__init__(for_class)
        self.storage = {}

    def _read_all(self):
        return list(self.storage.values())

    def _create(self, item):
        self.storage[item.get_id()] = item
        return item

    def _read(self, _id):
        if _id in self.storage:
            return self.storage[_id]

    def _update(self, item, insert=False):
        if item.get_id() not in self.storage:
            if insert:
                self.storage[item.get_id()] = item
        else:
            self.storage[item.get_id()] = item

    def _delete(self, _id):
        if _id in self.storage:
            copied_item = deepcopy(self.storage[_id])
            del self.storage[_id]
            return copied_item

    def _delete_all(self, *args, **kwargs):
        pass




import asyncio
import datetime
import re
from typing import List, Callable
from uuid import uuid4

from . import serializer
from .event_queue import EventQueue
from .events.error_response import ErrorResponse
from .events.event import Event
from .executor import ExecutorThread
from .exit_state import ExitState
from .handler.error_handler import ExceptionHandler
from .handler.handler_bootstrapper import HandlerBootstrapper
from .repository.repository_factory import AbstractRepositoryFactory
from .subscriber import Subscriber, SubscriberType
from .topic.topic_factory import TopicFactory


class EventDispatcher:
    def __init__(self, event_queue: EventQueue,
                 correlation_store,
                 thread_store,
                 client_store,
                 exception_handler: ExceptionHandler,
                 topic_factory: TopicFactory,
                 repository_factory: AbstractRepositoryFactory,
                 handler_bootstrapper: HandlerBootstrapper,
                 exit_state: ExitState,
                 handler_map=None,
                 protected_check: 'Callable[[Event], bool]' = None,
                 public_check: 'Callable[[Event], bool]' = None):
        self.event_queue = event_queue
        self.correlation_store = correlation_store
        self.thread_store = thread_store
        self.client_store = client_store
        self.exception_handler = exception_handler
        self.topic_factory = topic_factory
        self.repository_factory = repository_factory
        self.handler_map = handler_map if handler_map is not None else {}
        self.halted_for = 0
        self.handler_bootstrapper = handler_bootstrapper
        self.exit_state = exit_state
        self.protected_check = protected_check
        self.public_check = public_check
        self._init_handler_subscriptions()
        self.number_of_handled_events = 0
        self.number_of_outgoing_events = 0
        self.handle_times = []

    def _init_handler_subscriptions(self):
        handlers = self.handler_bootstrapper.get_handlers()
        for handler in handlers:
            if handler.subscribes_for():
                service_address = f"isc://{'-'.join(re.findall(r'[A-Z][a-z]+', handler.__name__)).lower()}"
                subscriber = Subscriber(
                    SubscriberType.SERVICE, service_address)
                for topic_name in handler.subscribes_for():
                    topic = self.topic_factory.get_topic(topic_name)
                    topic.add_subscriber(subscriber)

    def execute_on_existing_instance(self, event, service_name):
        actual_service_name = re.findall('(.*)-\d+$', service_name)[0]
        assert actual_service_name in self.handler_map, f'Actual service not found: {actual_service_name}'
        assert service_name in self.handler_map[
            actual_service_name], f'Service instance not found {service_name}'

        self.handler_map[actual_service_name][service_name].process_event(
            event)

    def thread_stopped(self, service_name, handler_name):
        actual_service_name = service_name
        assert actual_service_name in self.handler_map, f'Such service was never instansiated {actual_service_name}'
        assert handler_name in self.handler_map[
            actual_service_name], f'Service instance not found {handler_name}'
        assert self.handler_map[actual_service_name][
                   handler_name].handler_instance.running is False, f'Service still running, cannot stop {handler_name}'
        del self.handler_map[actual_service_name][handler_name]

    def instantiate_new_handler_and_execute(self, service_name, event, client_id):
        if service_name not in self.handler_map:
            self.handler_map[service_name] = {}

        instance_count = str(uuid4())
        handler_name = f'{service_name}-{instance_count}'

        handler_class = self.handler_bootstrapper.get_handler_class_by_service_name(service_name)

        if handler_class:
            executor_thread = ExecutorThread(
                handler_class,
                event.correlation_id,
                self.event_queue,
                self.exception_handler,
                self.topic_factory,
                self.repository_factory,
                handler_name,
                service_name,
                self.thread_stopped,
                public_check=self.public_check,
                protected_check=self.protected_check)

            self.thread_store.add(
                client_id,
                event.correlation_id,
                executor_thread)
            self.handler_map[service_name][handler_name] = executor_thread

            executor_thread.process_event(event)
            executor_thread.start()
        else:
            self.exception_handler.handle_exception(
                Exception(f'No handler for type {event.get_type()}'), event)

    async def dispatch(self):
        while not self.exit_state.should_exit:
            try:
                if len(self.event_queue) == 0:
                    await asyncio.sleep(0)
                else:
                    event = self.event_queue.get_event()
                    self.number_of_handled_events += 1

                    correlation_id = event.correlation_id
                    client_id = self.correlation_store.get(correlation_id)
                    destination = event.destination

                    if not destination:
                        # it's an incoming events from the client  => it should have a type,
                        # and should be a notification or request
                        # => destination can be deducted from request/notification type
                        # => it's should be directed to a service => isc://

                        handler_class = self.handler_bootstrapper.get_handler_for(event.get_type())
                        if not handler_class:
                            if issubclass(type(event), ErrorResponse):
                                print('Failed to dispatch ErrorResponse:', event)
                                continue
                            self.exception_handler.handle_exception(
                                Exception(f'> No handler for type {event.get_type()}'), event, use_return=True)
                            continue

                        service_name = self.handler_bootstrapper.get_service_name_from_class(
                            handler_class).strip()
                        destination = f'isc://{service_name}'
                        event.destination = destination

                        await self.dispatch_event_to_handler(client_id, event, handler_class, service_name)

                        continue
                    if destination.startswith('isc://'):
                        service_name = destination.split('//')[1].strip()

                        handler_class = self.handler_bootstrapper.get_handler_class_by_service_name(service_name)

                        await self.dispatch_event_to_handler(client_id, event, handler_class, service_name)

                    elif destination.startswith('remote://websocket/'):
                        client_id = re.findall(
                            r'[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}', destination)[0]
                        client = self.client_store.get(client_id)
                        
                        await client.websocket.send(serializer.serialize(event))
                        self.number_of_outgoing_events += 1
                        handling_time = (datetime.datetime.now() - datetime.datetime.fromisoformat(
                            event.system_entry)).microseconds / 1000
                        self.handle_times.append(handling_time)
                        self.handle_times = self.handle_times[-1000:]
            except Exception as dispatcher_e:
                print(dispatcher_e)

    async def dispatch_event_to_handler(self, client_id, event, handler_class, service_name):
        if handler_class:
            active_handlers = list(
                self.handler_map[service_name].values() if service_name in self.handler_map.keys() else [])
            available_handlers: List[ExecutorThread] = []
            for active_handler in active_handlers:
                key_readers = active_handler.handler_instance.key_readers
                for key_reader in key_readers:
                    try:
                        if key_reader(event):
                            available_handlers.append(active_handler)
                    except KeyError as key_exception:
                        print(
                            f'KeyReader exception: {key_exception} for keyreader in '
                            f'{active_handler.service_name}')
            if len(available_handlers) != 0:
                for available_handler in available_handlers:
                    available_handler.process_event(event)
            else:
                if event.get_type() in handler_class.started_by():
                    self.instantiate_new_handler_and_execute(service_name, event, client_id)
                else:
                    if event.retry_count < 5 and hasattr(event, 'get_type'):
                        if event.get_type().startswith('domain.'):
                            event.retry_count += 1
                            self.event_queue.add_event(event)
                            return

                    self.exception_handler.handle_exception(
                        Exception(
                            f'Handler type is found for {event.get_type()}, '
                            f'but there is no instance running, and the handler isn\'t started by it'),
                        event, use_return=True)
        else:
            self.exception_handler.handle_exception(
                Exception(f'No handler for type {event.get_type()}'), event, use_return=True)



from ..repository.storable import Storable


class AbstractRepository:
    def __init__(self, for_class):
        assert issubclass(
            for_class, Storable), 'for_class must extend Storable type'
        self._class = for_class

    def create(self, *args, **kwargs):
        self.check_type(args[0])
        return self._create(*args, **kwargs)

    def read(self, *args, **kwargs):
        return self._read(*args, **kwargs)

    def update(self, *args, **kwargs):
        self.check_type(args[0])
        return self._update(*args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._delete(*args, **kwargs)

    def read_all(self, *args, **kwargs):
        return self._read_all(*args, **kwargs)

    def delete_all(self, *args, **kwargs):
        return self._delete_all(*args, **kwargs)

    def _create(self, *args, **kwargs):
        raise NotImplementedError()

    def _read(self, *args, **kwargs):
        raise NotImplementedError()

    def _update(self, *args, **kwargs):
        raise NotImplementedError()

    def _delete(self, *args, **kwargs):
        raise NotImplementedError()

    def _read_all(self, *args, **kwargs):
        raise NotImplementedError()

    def _delete_all(self, *args, **kwargs):
        raise NotImplementedError()

    def check_type(self, item):
        assert issubclass(
            type(item), self._class), f'Item must be of type {self._class}'



from .notification import Notification


class ClientDisconnectedNotification(Notification):
    def __init__(self, client_id: str):
        super().__init__(payload={
            'client_id': client_id,
        })

    @staticmethod
    def get_type():
        return 'notification.client_disconnected'



from pathlib import Path

project_name: str = input('Name of project:')

main_template = '''
import os
from typing import Type, List

import pymongo
from dotenv import load_dotenv

from autumn import ExceptionHandler, Handler, HandlerBootstrapper, RepositoryExtension, MongoRepositoryFactory, \
    ExtensionBootstrapper
from autumn.event_dispatcher import EventDispatcher
from autumn.event_queue import EventQueue
from autumn.exit_state import ExitState
from autumn.store.client_store import ClientStore
from autumn.store.correlation_store import CorrelationStore
from autumn.store.thread_store import ThreadStore
from autumn.topic.topic_factory import TopicFactory
from autumn.websocket_server import WebsocketServer

handlers: 'List[Type[Handler]]' = []

extensions: 'List[Type[RepositoryExtension]]' = []

if __name__ == '__main__':
    load_dotenv('./.env')

    mongo_client = pymongo.MongoClient(f'mongodb://{os.environ["MONGODB_SERVER"]}:27017/')

    correlation_store = CorrelationStore()
    client_store = ClientStore()
    thread_store = ThreadStore()
    event_queue = EventQueue()
    handler_map = {}

    exception_handler = ExceptionHandler(event_queue)
    topic_factory = TopicFactory(event_queue)
    handler_bootstrapper = HandlerBootstrapper(handlers)
    repository_extension_bootstrapper = ExtensionBootstrapper(extensions)
    repository_factory = MongoRepositoryFactory(mongo_client, repository_extension_bootstrapper)

    exit_state = ExitState(topic_factory.get_topic('server_notifications'), client_store, thread_store, event_queue)

    event_dispatcher = EventDispatcher(event_queue, correlation_store, thread_store, client_store, exception_handler,
                                       topic_factory, repository_factory, handler_bootstrapper, exit_state,
                                       handler_map=handler_map)

    websocketServer = WebsocketServer(correlation_store, event_dispatcher, client_store, event_queue, exception_handler,
                                      topic_factory, exit_state)

    websocketServer.run(os.environ['HOST'], int(os.environ['PORT']))

'''
Path(f'./{project_name}').mkdir(parents=True)
Path(f'./{project_name}/contract').mkdir(parents=True)
Path(f'./{project_name}/contract/response').mkdir(parents=True)
Path(f'./{project_name}/contract/notification').mkdir(parents=True)
Path(f'./{project_name}/repository/repository_extension').mkdir(parents=True)
Path(f'./{project_name}/handler').mkdir(parents=True)

with open(f'./{project_name}/main.py', 'w') as f:
    f.write(main_template)

print(f'Created folder {project_name} with necessary folder structure!')

import socket

valid = False

while not valid:
    try:
        ip = input('The server should run on address: ')
        socket.inet_aton(ip)
        valid = True
    except socket.error as e:
        print(f'{ip} is not a valid address!')

valid = False
while not valid:
    try:
        port = int(input('The server must run on port: '))
        valid = True
    except ValueError as e:
        print(f'{port} is not a valid port!')

valid = False
while not valid:
    try:
        address = input('The mongodb server should run on address: ')
        socket.inet_aton(address)
        valid = True
    except socket.error as e:
        print(f'{address} is not a valid address!')

dot_env_file = f'MONGODB_SERVER={address}\nHOST={ip}\nPORT={port}\n'

with open(f'./{project_name}/.env', 'w') as f:
    f.write(dot_env_file)

print(f'Done! You can run the server by typing python3 main.py in the {project_name} folder!')





import os
from threading import Thread, active_count, enumerate
from time import sleep
import click

from .event_dispatcher import EventDispatcher
from .exit_state import ExitState
from .store.client_store import ClientStore
from .store.thread_store import ThreadStore
from .websocket_server import WebsocketServer


class InfoDisplay(Thread):
    def __init__(self, exit_state: ExitState, client_store: ClientStore, thread_store: ThreadStore, handler_map,
                 event_dispatcher: EventDispatcher, websocket_server: WebsocketServer):
        super(InfoDisplay, self).__init__()
        self.exit_state: ExitState = exit_state
        self.client_store = client_store
        self.thread_store = thread_store
        self.handler_map = handler_map
        self.event_dispatcher = event_dispatcher
        self.websocket_server = websocket_server

    def run(self) -> None:
        os.environ['TERM'] = 'xterm-256color'
        while not self.exit_state.should_exit:
            click.clear()
            number_of_dispatched_events = len(self.event_dispatcher.handle_times)
            avg = 0 if number_of_dispatched_events == 0 else sum(
                self.event_dispatcher.handle_times) / number_of_dispatched_events
            print(f'Currently running threads: {active_count()}\t\t'
                  f'Connected clients: {len(self.client_store.keys())}\t\t\t'
                  f'Running handlers: {len(self.handler_map.keys())}')
            print(f'Number of dispatched events: {self.event_dispatcher.number_of_handled_events}\t\t'
                  f'Number of incoming events: {self.websocket_server.number_of_incoming_events}\t\t'
                  f'Number of outgoing events: {self.event_dispatcher.number_of_outgoing_events}\t\t')
            print(f'Average handling delta: {avg}')
            
            print('Running threads:')
            for thread in enumerate():
                print(type(thread), end=' ## ')
            print()
            sleep(1)

        click.clear()
        print(f'Gracefully shutting down...')



from threading import Lock


class Store:
    def __init__(self):
        self.lock = Lock()

    def add(self, *args, **kwargs):
        with self.lock:
            return self._add(*args, **kwargs)

    def remove(self, *args, **kwargs):
        with self.lock:
            return self._remove(*args, **kwargs)

    def get(self, *args, **kwargs):
        with self.lock:
            return self._get(*args, **kwargs)

    def contains(self, *args, **kwargs):
        with self.lock:
            return self._contains(*args, **kwargs)

    def _add(self, *args, **kwargs):
        raise NotImplementedError()

    def _remove(self, *args, **kwargs):
        raise NotImplementedError()

    def _get(self, *args, **kwargs):
        raise NotImplementedError()

    def _contains(self, *args, **kwargs):
        raise NotImplementedError()



from datetime import datetime

from pymongo.collection import Collection

from ..repository.repository import AbstractRepository


class MongoRepository(AbstractRepository):
    def __init__(self, for_class, collection):
        super().__init__(for_class)
        self.for_class = for_class
        self.collection: Collection = collection

    def _create(self, item):
        return self.collection.insert_one(self.create_storable_format(item), {"last_modified": datetime.utcnow()})

    def _read(self, item_id):
        return self.create_from_storable_format(self.collection.find_one({'_id': item_id}))

    def _update(self, item, insert=False):
        return self.collection.update({'_id': item.get_id()}, self.create_storable_format(item), insert)

    def _delete(self, item_id):
        return self.collection.delete_one({'_id': item_id})

    def _read_all(self, *args, **kwargs):
        return list(map(lambda item: self.create_from_storable_format(item), self.collection.find({})))

    def _delete_all(self, *args, **kwargs):
        return self.collection.delete_many({})

    def create_storable_format(self, item):
        return dict(
            type=self.fullname(),
            payload=item.to_dict(),
            _id=item.get_id()
        )

    def create_from_storable_format(self, dct):
        if not dct:
            return None

        payload = dct['payload']
        item = self.for_class.from_dict(payload)
        if item.id == item.get_id() and item.get_id() != dct['_id']:
            item.id = dct['_id']

        return item

    def fullname(self):
        module = self.for_class.__module__

        if not module or module == str.__class__.__module__:
            return self.for_class.__name__
        else:
            return f'{module}.{self.for_class.__name__}'



from .event_dispatcher import EventDispatcher
from .events import Notification, Response, DomainEvent, Request
from .events.client_disconnected import ClientDisconnectedNotification
from .events.server_disconnect import DisconnectedNotification
from .events.error_response import ErrorResponse
from .events.shutdown_requested import ShutdownRequestedNotification
from .handler.decorators import *
from .handler.handler import Handler
from .handler.handler_bootstrapper import HandlerBootstrapper
from .handler.error_handler import ExceptionHandler
from .event_queue import EventQueue
from .exit_state import ExitState
from .executor import ExecutorThread
from .websocket_server import WebsocketServer
from .state_machine.state import State
from .state_machine.state_configurator import StateConfigurator
from .state_machine.state_machine import StateMachine
from .state_machine.state_machine_builder import StateMachineBuilder
from .repository.memory_repository_factory import MemoryRepositoryFactory
from .repository.memory_repository import MemoryRepository
from .repository.mongo_repository import MongoRepository
from .repository.mongo_repository_factory import  MongoRepositoryFactory
from .repository.repository import AbstractRepository
from .repository.storable import Storable
from .repository.repository_extension.decorators import *
from .repository.repository_extension.extension_bootstapper import ExtensionBootstrapper
from .repository.repository_extension.extension import RepositoryExtension


from typing import Callable, Dict

from ..repository.storable import Storable


class State:
    def __init__(self, name, context=None):
        self.name = name
        self.next_states = {}
        self.context: Storable = context
        self.transition_listeners: 'Dict[State, Callable[[Storable], None]]' = {}

    def add_allowed_transition(self, trigger, next_state):
        assert issubclass(type(next_state), State), "Only objects dervied from state must be submitted"
        assert trigger not in self.next_states, 'Only one state transition permitted per trigger'
        self.next_states[trigger] = next_state

    def is_transition_allowed(self, trigger):
        return trigger in self.next_states.keys()

    def get_transition_state(self, trigger) -> 'State':
        assert self.is_transition_allowed(trigger), f'Transition is not allowed from {self.name} with {trigger}'
        next_state = self.next_states[trigger]
        print(f'Transitioning from {self.name} to {next_state.name}')
        if next_state in self.transition_listeners:
            executable: Callable[[Storable], None] = self.transition_listeners[next_state]
            executable(self.context)

        return self.next_states[trigger]

    def on_transition_to(self, to_state: 'State', a_callable: Callable[[Storable], None]):
        assert to_state in self.next_states.values(), f'Transition from {self.name} to {to_state.name} is not allowed!' \
                                                      f'awailable states: {[state.name for state in self.next_states.values()]}'
        self.transition_listeners[to_state] = a_callable

    def __str__(self):
        return self.name

    def to_dict(self):
        allowed_states = dict([(trigger, state.name) for trigger, state in self.next_states.items()])
        return {
            'name': self.name,
            'allowed_stated': allowed_states
        }



import json
from copy import deepcopy
from datetime import datetime

from .events.request import Request


def get_correlation_id(obj, correlation_id):
    correlation_id = obj['correlation_id'] if 'correlation_id' in obj else correlation_id
    correlation_id = obj['correlationId'] if 'correlationId' in obj else correlation_id

    return correlation_id


def deserialize(string, correlation_id, client_id):
    obj = json.loads(string)

    if 'type' not in obj:
        raise Exception('This request format was not recognized by the system')

    type_ = obj['type']
    payload = obj['payload'] if 'payload' in obj else {}
    headers = obj['headers'] if 'headers' in obj else {}

    request = Request(type_, payload=payload, headers=headers)
    request.correlation_id = get_correlation_id(obj, correlation_id)
    request.client_id = client_id
    request.system_entry = str(datetime.now())

    return request


def serialize(obj):
    dct = _recursive_clear_none(deepcopy(obj.__dict__))
    dct['type'] = obj.get_type()

    return json.dumps(dct, indent=4)


def _recursive_clear_none(obj):
    if type(obj) == dict:
        new_obj = {}
        for key in obj.keys():
            result = _recursive_clear_none(obj[key])
            if result is not None:
                new_obj[transform_to_camel_case(key)] = result
        return new_obj

    return obj


def transform_to_camel_case(key):
    chunks = key.split('_')
    return chunks[0] + ''.join([chunk.capitalize() for chunk in chunks[1:]])



import setuptools

setuptools.setup(
    name='autumn-boot',
    version='1.0.1',
    author='Ferencz Albert',
    author_email='split@aferencz.xyz',
    description='A framework to build event-driven backend solutions',
    long_description='',
    url='https://bitbucket.org/AlphaSpliT/autumn',
    packages=setuptools.find_packages(where='.'),
    python_requires='>=3.6',
    install_requires=[
        'asyncio',
        'websockets',
        'python-dotenv',
        'click',
        'pymongo',
    ],
    setup_requires=['wheel']
)



from typing import Type

from pymongo.collection import Collection


class RepositoryExtension:
    for_class: Type
    collection: Collection

    def create_storable_format(self, item):
        pass

    def create_from_storable_format(self, dct):
        pass



from .store import Store


class ThreadStore(Store):
    def __init__(self):
        super().__init__()
        self.map = {}

    def _add(self, client_id, correlation_id, executor_thread):
        if client_id in self.map:
            self.map[client_id][correlation_id] = executor_thread
        else:
            self.map[client_id] = {}
            self.map[client_id][correlation_id] = executor_thread

    def _get(self, client_id):
        return self.map[client_id] if client_id in self.map else {}

    def _contains(self, client_id, correlation_id=None):
        if not correlation_id:
            return client_id in self.map
        else:
            return client_id in self.map and correlation_id in self.map[client_id]

    def _remove(self, client_id, correlation_id=None):
        if client_id in self.map:
            if correlation_id and correlation_id in self.map[client_id]:
                del self.map[client_id][correlation_id]
            elif not correlation_id:
                del self.map[client_id]



def event_handler(event_to_handle):
    def sub_decorator(method):
        assert type(event_to_handle) == str, "Event to handle must be of type string"
        method.handles = event_to_handle
        return method

    return sub_decorator


def starts_handler(method):
    method.starts_handler = True

    return method


def tear_down(method):
    method.is_teardown = True

    return method


def protected(method):
    method.protected = True
    return method


def only_public(method):
    method.only_public = True
    return method



from ..events import Response


class ErrorResponse(Response):
    def __init__(self, message='', status=500):
        super().__init__(status=status, message=message)

    @staticmethod
    def get_type():
        return 'response.error'



from enum import Enum


class Subscriber:
    def __init__(self, subscriber_type: 'SubscriberType', address):
        self.type = subscriber_type
        self.address: str = address

    def __str__(self):
        s = ''
        for prop, value in vars(self).items():
            s += f'{str(prop)}={str(value)} '

        return f'<{self.__class__.__name__} {s} />'

    def __eq__(self, other: 'Subscriber'):
        return self.type == other.type and self.address == other.address

    def __hash__(self):
        return self.type.__hash__() + self.address.__hash__()

    @classmethod
    def client_from_id(cls, client_id):
        return cls(SubscriberType.CLIENT, f'remote://websocket/{client_id}')


class SubscriberType(Enum):
    CLIENT = 1
    SERVICE = 2
